import type { DataXY } from 'cheminfo-types';
import { toBeDeepCloseTo, toMatchCloseTo } from 'jest-matcher-deep-close-to';
import { getShape1D, Shape1D } from 'ml-peak-shape-generator';
import noisyBigPeakSmallPeak from 'nmr-xy-testdata/data/noisy/noisyBigPeakSmallPeak.json';
import tripletQuadruplet from 'nmr-xy-testdata/data/pure/d1-2_j7.json';
import { generateSpectrum } from 'spectrum-generator';

import { xyAutoPeaksPicking } from '../xyAutoPeaksPicking';

expect.extend({ toBeDeepCloseTo, toMatchCloseTo });

describe('xyAutoPeaksPicking', () => {
  it('a triplet and a quadruple', () => {
    let options = {};

    let peaks = xyAutoPeaksPicking(tripletQuadruplet, options);

    const widthToFWHM = getShape1D({ kind: 'gaussian' }).widthToFWHM;
    expect(peaks).toHaveLength(7);
    expect(peaks[1].x).toBeDeepCloseTo(0.999831, 3);
    expect(peaks[1].y / 100).toBeDeepCloseTo(14846602893.68, 1);
    expect(peaks[1].width).toBeDeepCloseTo(widthToFWHM(0.0021514892578125), 3);
  });
  it('mixed spectrum with small and big peaks', () => {
    let options = {
      thresholdFactor: 3,
      compile: false,
      clean: false,
      optimize: false,
      minMaxRatio: 0.01,
      integralType: 'sum',
      nH: 3,
      frequencyCluster: 16,
      widthFactor: 4,
      smoothY: false,
      broadWidth: 0.2,
      shape: { kind: 'lorentzian' } as Shape1D,
      broadRatio: 0,
    };

    let peaks = xyAutoPeaksPicking(noisyBigPeakSmallPeak, options);
    const widthToFWHM = getShape1D({ kind: 'lorentzian' }).widthToFWHM;

    const expectedResult: { [key: string]: any }[] = [
      { x: 2, y: 6.2683, width: 0.4, fwhm: widthToFWHM(0.4) },
      { x: 8, y: 318.611, width: 0.4, fwhm: widthToFWHM(0.4) },
    ];

    expectedResult.forEach((expected, i) => {
      let peak = peaks[i];
      for (let key in expected) {
        //@ts-expect-error
        if (typeof peak[key] === 'object') {
          //@ts-expect-error
          expect(peak[key]).toMatchCloseTo(expected[key], 2);
        } else {
          //@ts-expect-error
          expect(peak[key]).toBeCloseTo(expected[key], 2);
        }
      }
    });
  });

  it('negative spectrum', () => {
    let y = tripletQuadruplet.y;
    for (let i = 0; i < y.length; i++) {
      y[i] *= -1;
    }
    tripletQuadruplet.y = y;
    let options = { lookNegative: true };
    let peaks = xyAutoPeaksPicking(tripletQuadruplet, options);
    expect(peaks).toHaveLength(7);
    expect(peaks[1].x).toBeDeepCloseTo(0.999831, 3);
    expect(peaks[1].y / 100).toBeDeepCloseTo(-14846602893.68, 1);
    expect(peaks[1].width).toBeDeepCloseTo(0.0021514892578125, 3);
  });

  it('negative and positive data', () => {
    const peaks = [
      { x: -0.5, y: 1, width: 0.05 },
      { x: 0.5, y: -1, width: 0.05 },
    ];

    const data: DataXY = generateSpectrum(peaks, {
      generator: {
        from: -1,
        to: 1,
        nbPoints: 101,
      },
      peakOptions: {
        factor: 6,
      },
    });

    const peakList = xyAutoPeaksPicking(data, {
      lookNegative: true,
      smoothY: false,
      noiseLevel: 0,
    });

    let expected = [
      {
        x: -0.5,
        y: 1,
        ddY: -259.83290100626783,
        fwhm: 0.09419280180123805,
        width: 0.08,
        shape: {
          kind: 'gaussian',
        },
        index: 25,
        inflectionPoints: {
          from: { index: 23, x: -0.54 },
          to: { index: 27, x: -0.46 },
        },
      },
      {
        x: 0.5,
        y: -1,
        ddY: 259.83290100626783,
        fwhm: 0.09419280180123805,
        width: 0.08,
        shape: {
          kind: 'gaussian',
        },
        index: 75,
        inflectionPoints: {
          from: { index: 73, x: 0.46 },
          to: { index: 77, x: 0.54 },
        },
      },
    ];
    expect(peakList).toBeDeepCloseTo(expected);
  });
});
